//--------------------------------------------------------------
//                     TRAITEMENT D'IMAGE                      
//--------------------------------------------------------------


// Prototypes traitement d'image
struct palettetype pal;
void Affiche_Image     (unsigned char ***, short    , short            , short, short, short);
void Seuillage_Image   (unsigned char ***, short    , unsigned char ***, short, short, short, short, short, short);
void Lissage_Image     (unsigned char ***, short    , unsigned char ***, short, short, short, short, short);
void Histogramme_Image (unsigned char ***, float ** , short, short     , short, short, short, short, short, short, short, short);


//--------------------------------------------------------------
// AFFICHAGE D'UNE IMAGE
//--------------------------------------------------------------
// Entres : Pointeur sur le tenseur image d'unsigned char
//				 Indice de dbut de ligne
//				 Indice de fin de ligne
//				 Indice de dbut de colonne
//				 Indice de fin de colonne
//           Indice de couleur  afficher :
//                                      ROUGE = 0
//													 VERT  = 1
//													 BLEU  = 2
//													 GRIS (moyenne RGB) = 3
// Sortie : void
//--------------------------------------------------------------
void Affiche_Image(unsigned char ***imageRGB,
						 short debutligne,
						 short finligne,
						 short debutcolonne,
						 short fincolonne,
						 short couleur)
 {
  int i, j;
  int errorcode;
  struct palettetype palette;

  errorcode = graphresult ();
	if ( errorcode != grOk)
	{
	 printf ("Erreur graphique : %s\n", grapherrormsg (errorcode));
	 printf ("Pressez une touche...\n");
	 return ;
	}
  getpalette (&palette);
  for (i = 0; i <= palette.size; i++ )
	 setrgbpalette (palette.colors[i], i*4,i*4,i*4);
  for (i = debutligne; i < finligne; i++ )
	 for (j = debutcolonne; j < fincolonne; j++ )
		putpixel (i, j, imageRGB[couleur][i][j]/16);
}


//--------------------------------------------------------------
// SEUILLAGE D'UNE IMAGE
//--------------------------------------------------------------
// Entres : Pointeur sur le tenseur de l'image  seuiller
//           Indice de la couleur  seuiller
//           Pointeur sur le tenseur de l'image seuille
//				 Indice de dbut de ligne
//				 Indice de fin de ligne
//				 Indice de dbut de colonne
//				 Indice de fin de colonne
//				 Seuil de seuillage infrieur
//           Seuil de seuillage suprieur
// Sortie : void
//--------------------------------------------------------------
// Seuillage d'une image entre 2 seuils de luminance :
// Affichage des pixels dont la valeur de luminance de situe dans
// l'intervalle form par les 2 seuils
//--------------------------------------------------------------
void Seuillage_Image(unsigned char ***Image,
							short couleur,
							unsigned char ***Image_Seuillee,
							short Debut_Ligne ,
							short Fin_Ligne,
							short Debut_Colonne,
							short Fin_Colonne,
							short Seuil_Inferieur, // De luminance (entre 0 et 255)
							short Seuil_Superieur)
 {
  short i, j;
  for (i = Debut_Ligne; i <= Fin_Ligne; i ++)
	 for (j = Debut_Colonne; j <= Fin_Colonne; j ++)
	  {
		if ( (Image[couleur][i][j] <= Seuil_Superieur) && (Image[couleur][i][j] >= Seuil_Inferieur) )
		  Image_Seuillee[couleur][i][j] = Image[couleur][i][j];
		else
		  Image_Seuillee[couleur][i][j] = 0;
	  }
 }

//--------------------------------------------------------------
// LISSAGE D'UNE IMAGE
//--------------------------------------------------------------
// Entres : Pointeur sur le tenseur de l'image  lisser
//           Indice de la couleur  seuiller
//           Pointeur sur le tenseur de l'image lisse
//				 Indice de dbut de ligne
//				 Indice de fin de ligne
//				 Indice de dbut de colonne
//				 Indice de fin de colonne
//				 Largeur du carr de seuillage
// Sortie : void
//--------------------------------------------------------------
// Lissage d'une image : Moyenne des luminances des pixels
// voisins sur un carr de 'Largeur_Seuillage' de cot
//--------------------------------------------------------------

void Lissage_Image(unsigned char ***Image,
						 short couleur,
						 unsigned char ***Image_Lissee,
						 short Debut_Ligne ,
						 short Fin_Ligne,
						 short Debut_Colonne,
						 short Fin_Colonne,
						 short Largeur_Seuillage)
 {
  short i, j, k, l, nbpixels, Somme_Luminance;

  for (i = Debut_Ligne; i <= Fin_Ligne; i ++)
	 for (j = Debut_Colonne; j <= Fin_Colonne; j ++)
	  {
		nbpixels = 0;
		Somme_Luminance = 0;
		for (k = -(Largeur_Seuillage-1)/2; k <= (Largeur_Seuillage-1)/2; k ++)
			for (l = -(Largeur_Seuillage-1)/2; l <= (Largeur_Seuillage-1)/2; l ++)
			  {
				if ( ((i+k) >= 0) && ((j+l) >= 0) && ((i+k) <= 255) && ((j+l) <= 255) )
				 {
				  nbpixels ++;
				  Somme_Luminance += Image[couleur][i+k][j+l];
				 }
			  }
		Image_Lissee[couleur][i][j] = Somme_Luminance / nbpixels;
	  }
 }


//--------------------------------------------------------------
// CALCULE ET TRACE DE L'HISTOGRAMME D'UNE IMAGE
//--------------------------------------------------------------
// Entres : Pointeur sur le tenseur de l'image  lisser
//           Indice de la couleur dont on veut tracer l'histogramme
//				 Indice de dbut de ligne
//				 Indice de fin de ligne
//				 Indice de dbut de colonne
//				 Indice de fin de colonne
//           Gauche de la fentre de trac de l'histogramme
//           Haut de la fentre de trac de l'histogramme
//           Droite de la fentre de trac de l'histogramme
//           Bas de la fentre de trac de l'histogramme
// Sortie : void
//--------------------------------------------------------------
// Histogramme d'une image : Nombre de pixels ayant la mme luminance
// Sur l'axe horizontal : luminance : de 0  255
// Sur l'axe vertical : nombre de pixels de l'image ayant la luminance
// indique sur l'axe horizontal
//--------------------------------------------------------------

void Histogramme_Image(unsigned char ***Image,
                       float **Histogramme_Image,
							  short couleur,
							  short Debut_Ligne,
							  short Fin_Ligne,
							  short Debut_Colonne,
							  short Fin_Colonne,
							  short gauche,
							  short haut,
							  short droite,
							  short bas, 
							  short couleur_affichage)
 {
  short largeur = (droite-gauche), hauteur = (bas-haut), i, j;
  unsigned char pixel;

  setviewport (gauche, haut,droite, bas, 1);
  clearviewport ();
  setfillstyle (SOLID_FILL, couleur_affichage);
  // Initialisation de l'histogramme  0
  for (i = 0; i <= 256; i ++)
	 Histogramme_Image[couleur][i] = 0.;
  // Calcul de l'histogramme et recherche du maximum de population
  for (i = Debut_Ligne; i <= Fin_Ligne; i ++)
	 for (j = Debut_Colonne; j <= Fin_Colonne; j ++)
	  {
		pixel = Image[couleur][i][j];
		Histogramme_Image[couleur][(short)pixel] ++;
		if (Histogramme_Image[couleur][256] <= Histogramme_Image[couleur][(short)pixel])
		  Histogramme_Image[couleur][256]  = Histogramme_Image[couleur][(short)pixel]; // Maximum de population
	  }
  // Trac de l'histogramme
  for (i = 0; i <= 255; i ++)
	 bar (i*(float)largeur/256.,hauteur*(1-(Histogramme_Image[couleur][i]/
	 Histogramme_Image[couleur][256])),(i+1)*(largeur/256.),hauteur);
  setviewport (0, 0, getmaxx(), getmaxy(), 1);
  Liberation_Tenseur_FLOAT(Histogramme_Image,1, 0, 2, 0, 256);
 }
