#include "lzw.h"
#include "fbits.h"	// notre librairie d'entres-sorties de bits sur disque,
							// pour les constructeurs ifBits et ofBits et les fonc-
							// tions writeBits et readBits


// Constructeur (voir aussi lzwdico.h et lzwdico.cpp)
LZW::LZW( unsigned int nombre_bits_final )
	: NBITS_FINAL( nombre_bits_final ),
	  DICO( *new LZWDico( nombre_bits_final ) ),
	  PILE( *new Pile( LZWDico::maxval( nombre_bits_final ) ) )
{ }


// Destructeur
LZW::~LZW()
{
	delete & DICO ; delete & PILE ;
}


LZW::erreur LZW::compresse( const char * source, const char * cible )
//
// On compresse le fichier existant de nom source, vers un nouveau
// fichier devant tre nomm cible. On retourne le statut final sous
// la forme d'un type numr.
{
	// Ouvrir, si possible, le fichier source comme un flux d'octets,
	// et le fichier cible comme un flux de bits (notre classe ofBits)
	ifstream SOURCE( source, ios::binary ) ;
	ofBits 	CIBLE( cible, ios::binary ) ;
	if( ! SOURCE.good() )
		return fsource ;
	if( ! CIBLE.good() )
		return fcible ;

	// Laisser de la place pour l'en-tte au dbut du fichier cible
	ecrireEnTeteDeb( CIBLE ) ;
	LG_ORIG = 0 ;	CRC_ORIG.raz() ;

	// Ce qui suit est l'implmentation, avec nos outils et classes, de
	// l'algorythme LZW tel que dcrit par Mark Nelson ; voir son livre

	DICO.vider() ;

	// Lire le premier octet ; si source vide, mettre le code FIN_FICHIER,
	// sinon mettre caractre lu, puis incrmenter longueur et calculer crc
	int car = SOURCE.get() ;
	LZWEntree::code code = car == EOF ? LZWEntree::FIN_FICHIER : ( LZWEntree::code )car ;
	if( car != EOF )
	{
		++LG_ORIG ;
		CRC_ORIG.valeur( ( unsigned char )car ) ;
	}

	// Boucle de traitement des caractres suivants lus dans le source

	while( /* SOURCE.good() && CIBLE.good() && */ ( car = SOURCE.get() ) != EOF )
	// les tests en commentaires sont trs souhaitables thoriquement, hlas
	// ils cotent du temps : si source est 1 MO, on passe 1 million de fois !
	{
		// Pour chaque caractre lu, incrmenter longueur, et recalculer crc
		++LG_ORIG ;	CRC_ORIG.valeur( ( unsigned char )car ) ;

		// Recherche dans le dictionnaire
		LZWEntree::code index = DICO.index( code, car ) ;
		// Si l'on a trouv, le code de cette entre devient le code courrant
		if( DICO[ index ].leCode() != LZWEntree::RIEN )
			code = DICO[ index ].leCode() ;
		// Si l'on a pas trouv, on ajoute une entre au dictionnaire, en
		// mettant le code avec le nombre de bits actuel (car il variera)
		else
		{
			DICO[ index ].ajoute( DICO.codeSuivant(), code, (unsigned char)car ) ;
			CIBLE.writeBits( code, DICO.nbits() ) ;
			code = ( LZWEntree::code ) car ;

			// On demande au dictionnaire d'allouer un autre code libre ; celui
			// ci est mis dans la variable ancien_nombre_bits, passe comme une
			// rfrence. Il y a 2 cas particuliers : si la longueur actuelle du
			// code est sature, le dico l'augmente, et nous signale cela en re-
			// tournant pluslong (type numr) ; nous devons alors mettre un
			// code spcial pour prvenir le dcompresseur de faire la mme cho-
			// se, au mme moment ; d'autre part, si la longueur de code elle-
			// mme est arrive  son maximum, le dico se vide, et nous le si-
			// gnale en retournant troplong ; nous devons alors mettre pour le
			// dcompresseur le code spcial de vidage du dictionnaire.
			// Il importe de remarquer qu'en cas de changement de longueur du
			// code, le code spcial signalant ce fait est lui-mme mis avec
			// l'ancienne longueur.
			unsigned int ancien_nombre_bits ;
			switch( DICO.incrementer( ancien_nombre_bits ) )
			{
				case LZWDico::troplong : CIBLE.writeBits( LZWEntree::VIDER_DICO, ancien_nombre_bits ) ;
												 break ;
				case LZWDico::pluslong : CIBLE.writeBits( LZWEntree::ALLONGER_CODE, ancien_nombre_bits ) ;
			}
		}
	}
	// On crit le dernier octet, puis le code spcial de fin de fichier
	CIBLE.writeBits( code, DICO.nbits() ) ;
	CIBLE.writeBits( LZWEntree::FIN_FICHIER, DICO.nbits() ) ;
	// On revient sur l'en-tte, puisqu'on possde dsormais les informations
	// pour le remplir (longueur, crc, etc).
	return ecrireEnTeteFin( CIBLE ) ;
}


LZW::erreur LZW::decompresse( const char * source, const char * cible )
//
// Cette fois, source est le nom du fichier comprim,et cible celui du
// nouveau fichier dcompress  crer. On renvoie un statut d'erreur.
{
	ifBits 	SOURCE( source, ios::binary ) ;	// flux de lecture de bits
	ofstream	CIBLE( cible, ios::binary ) ;		// flux d'criture d'octets
	if( ! SOURCE.good() )
		return fsource ;
	if( ! CIBLE.good() )
		return fcible ;

	LG_ORIG = 0 ;	CRC_ORIG.raz() ;

	// On lit l'en-tte et l'on vrifie son intgrit
	erreur er = lireEnTete( SOURCE ) ;
	if( er != ras )
		return er ;

	// Algorythme de dcompression LZW (cf Mark Nelson)

	while( 1 /* SOURCE.good() && CIBLE.good() */ )
	// tests souhaitables mais coteux en temps
	{
		DICO.vider() ;	// survient aussi  rception du code de vidage
		LZWEntree::code ancien_code = 0 ;

		// Traitement du premier code lu (avec la longueur de code initiale)
		SOURCE.readBits( ancien_code, DICO.nbits() ) ;
		if( ancien_code == LZWEntree::FIN_FICHIER )
			return ras ;
		unsigned char car = ( unsigned char )ancien_code ;
		CIBLE.put( car ) ;	++LG_ORIG ;	 CRC_ORIG.valeur( car ) ;

		// Traitement des autres codes

		while( 1 /* SOURCE.good() && CIBLE.good() */ )
		// tests souhaitables mais coteux en temps
		{
			LZWEntree::code nouveau_code = 0 ;
			SOURCE.readBits( nouveau_code, DICO.nbits() ) ;

			// Fin du source ; on vrifie sa longueur, son
			// crc, par rapport aux donnes de l'en-tte
			if( nouveau_code == LZWEntree::FIN_FICHIER )
				return verifEnTete() ;

			// Traitement des codes spciaux mis par le compresseur :
			// allongement du code, ou vidage du dictionnaire. Il est
			// essentiel de faire exactement la mme chose que le com-
			// presseur, et au mme moment !
			if( nouveau_code == LZWEntree::VIDER_DICO )
				break ;
			if( nouveau_code == LZWEntree::ALLONGER_CODE )
			{
				DICO.allongerCode() ;
				continue ;
			}

			// Cas particulier de l'algo LZW : on reoit un code
			// qui n'est pas encore dans le dictionnaire ! Il faut
			// empiler le caractre dans la pile (cf Mark Nelson)
			if( nouveau_code >= DICO.codeSuivant() )	// cas spcial : code inconnu
			{
				PILE.push( car ) ;
				empilerChaine( ancien_code ) ;
			}
			else
				empilerChaine( nouveau_code ) ;

			// On dpile en mettant les caractres, et l'on met  jour le crc
			car = PILE.top() ;				// simple consultation sans dpilement
			while( ! PILE.isEmpty() )
			{
				unsigned char c = PILE.pop() ;
				CIBLE.put( c ) ;	++LG_ORIG ;	 CRC_ORIG.valeur( car ) ;
			}

			// On ajoute une entre au dictionnaire
			DICO[ DICO.codeSuivant() ].ajoute( LZWEntree::RIEN, ancien_code, car ) ;
			DICO.incrementer() ;
			ancien_code = nouveau_code ;
		}
	}
}


void LZW::empilerChaine( LZWEntree::code code_depart )
//
// On remonte l'arborescence du dictionnaire depuis code_depart,
// en empilant dans notre PILE les caractres rencontrs.
//
{
	LZWEntree::code code = code_depart ;

	while( code >= LZWDico::codeDepart() )
	{
		PILE.push( DICO[ code ].leCar() ) ;
		code = DICO[ code ].leParent() ;
	}
	PILE.push( ( unsigned char )code ) ;
}


#ifdef DEPANNAGE
void LZW::lister() const
//
// Vous voyez bien que vous n'tes pas seul  prouver
// des difficults de mise au point de temps  autre !
{
	DICO.lister() ;
}
#endif


LZW::erreur LZW::lireEnTete( ifBits & fbits )
//
// On lit et on vrifie l'en-tte, en renvoyant
// nos conclusions sous forme d'un type numr.
{
	// Lire l'en-tte comme un tableau d'octets
	fbits.read( en_tete.ucars, sizeof( en_tete.ucars ) ) ;

	// Le crc des informations de l'en-tte qui fut stock par le dcom-
	// presseur doit tre gal  celui que nous recalculons maintenant.
	if( en_tete.infos.crc_infos != CRCEnTete() )
		return crctete ;

	// La longueur de code initiale et maximale tant paramtrables,
	// il faut s'assurer que l'on dcompresse avec les mmes bornes
	// que lors de la compression : une dcompression 9  13 bits
	// aprs une compression 9  12 bits par exemple, donnerait vi-
	// demment du chinois (ou du franais, si vous tes chinois !).
	if( en_tete.infos.nbits_depart != 9 )
		return version ;
	if( en_tete.infos.nbits_limite != NBITS_FINAL )
		return version ;

	return ras ;	// ouf, c'est heureusement le cas le plus frquent !
}


LZW::erreur LZW::verifEnTete() const
//
// On compare la longueur et le crc recalculs du fichier
// avec leurs valeurs mmorises dans l'en-tte , et l'on
// renvoie le verdict sous forme d'un type numr.
{
	if( en_tete.infos.lg_fich_originel != LG_ORIG )
		return lgfich ;
	if( en_tete.infos.crc_fich_originel != CRC_ORIG.valeur() )
		return crcfich ;

	return ras ;
}


void LZW::ecrireEnTeteDeb( ofBits & fbits )
//
// Nous crivons squentiellement dans le flux cible en argument,
// donc il nous faut ds le dbut rserver la place ncessaire
// pour l'en-tte. Comme la plupart des informations sont indis-
// ponibles avant d'avoir lu entirement le source (longueurs,
// crc, etc), nous remplissons provisoirement avec des zros...
{
	en_tete.infos.crc_infos				= 0 ;	// sera mis  jour par ecrireEnTeteFin
	en_tete.infos.nbits_depart			= 9 ;
	en_tete.infos.nbits_limite			= NBITS_FINAL ;
	en_tete.infos.lg_fich_originel	= 0 ;	// sera mis  jour par ecrireEnTeteFin
	en_tete.infos.crc_fich_originel	= 0 ; // sera mis  jour par ecrireEnTeteFin
	en_tete.infos.lg_fich_comprime	= 0 ; // sera mis  jour par ecrireEnTeteFin

	fbits.write( en_tete.ucars, sizeof( en_tete.ucars ) ) ;
}


LZW::erreur LZW::ecrireEnTeteFin( ofBits & fbits )
//
// En fin de compression, nous disposons des informations  mmoriser dans
// l'en-tte ; donc nous y revenons (en dbut de fichier) pour les mettre 
// jour. Noter que ces informations d'en-tte font elles-mmes l'objet d'un
// crc mmoris, qui sera vrifi en tout premier lieu par le dcompresseur.
{
	fbits.flush() ;									// sage prcaution
	unsigned long lg_compr = fbits.tellp() ;	// longueur comprime

	// Remplir la structure d'en-tte avec les infos dsormais disponibles
	en_tete.infos.lg_fich_originel	= LG_ORIG ;
	en_tete.infos.crc_fich_originel	= CRC_ORIG.valeur() ;
	en_tete.infos.lg_fich_comprime	= lg_compr ;
	en_tete.infos.crc_infos				= CRCEnTete() ;	//  laisser en dernier !!
																		// (crc de ce qui prcde =
																		//  auto-vrif de l'en-tte)

	// Revenir sur l'en-tte en dbut de fichier, pour le r-crire
	fbits.seekp( 0 ) ;
	fbits.write( en_tete.ucars, sizeof( en_tete.ucars ) ) ;

	// Renvoie un avertissement si le fichier comprim est gal ou suprieur
	//  l'original (peut arriver si l'on comprime un fichier dj comprim)
	return lg_compr < LG_ORIG ? ras : taille ;
}


CRC32::crc LZW::CRCEnTete() const
//
// Calcule le crc des informations de l'en-tte, qui sera lui mme mmoris
// dans l'en-tte pass au dcompresseur. Voir aussi crc32.h et crc32.cpp.
{
	CRC32 C ;	// classe de calcul de crc, cf crc32.h et crc32.cpp
	C.valeur( en_tete.ucars+sizeof( en_tete.infos.crc_infos ), sizeof( en_tete.ucars ) - sizeof( en_tete.infos.crc_infos ) ) ;
	return C.valeur() ;	// consultation de crc calcul
}


const char * LZW::message( erreur er ) const
//
// Renvoie une chane de message d'erreur explicite,
// selon la valeur du code en argument (type numr)
{
	char * pt ;
	switch( er )
	{
		case ras			:	pt = "aucune erreur" ;
								break ;
		case taille		:	pt = "taille compresse suprieure ; original probablement compress" ;
								break ;
		case version  	: 	pt = "la compression fut effectue avec une autre version du programme" ;
								break ;
		case lgfich   	:  pt = "le fichier dcompress a une longueur diffrente de l'originel" ;
								break ;
		case crcfich  	: 	pt = "le crc du fichier dcompress est diffrent de l'originel" ;
								break ;
		case crctete	:  pt = "le crc de l'en-tte est erronn" ;
								break ;
		case fsource   :  pt = "impossible d'ouvrir le fichier source" ;
								break ;
		case fcible		:  pt = "impossible de crer le fichier cible" ;
								break ;
		default			:	pt = "erreur non rpertorie" ;	// ne devrait pas survenir
	}
	return ( const char * )pt ;	// tricher pour la bonne cause (expression STROUSTRUPienne !)
											// (j'ai assign pt dans le switch, donc je ne pouvais le d-
											//  clarer const ; toutefois, je dois tenir la promesse faite
											//  quand j'ai dclar renvoyer un const char *).
}
