#include "lzwdico.h"
#include <assert.h>
#include <iostream.h>

// Code spcial signifiant chec d'une recherche de caractre
const LZWEntree::code LZWEntree::RIEN 				= 256 ;
// (cette smantique du RIEN est courante : lorsque vous chouez
//  trouver un pointeur, vous renvoyez 0 ; lorsque EXCEL choue
//  trouver une valeur, il renvoie #N/A ; lorsque vous cherchez
// un index positif vous pouvez renvoyer -1 s'il est absent, etc ;
// Moi j'aime bien appeler les choses par leur nom : RIEN !!)
//
// Je rappelle que les codes 0  255 sont rservs pour les carac-
// tres ordinaires trouvs dans le source  comprimer.

// Codes spciaux mis  destination du dcompresseur
const LZWEntree::code LZWEntree::FIN_FICHIER    = 257 ;
const LZWEntree::code LZWEntree::ALLONGER_CODE  = 258 ;
const LZWEntree::code LZWEntree::IGNORER_DICO   = 259 ;	// ne sert pas
const LZWEntree::code LZWEntree::VIDER_DICO     = 260 ;

// Divers paramtres pour assurer l'volutivit du code
const LZWEntree::code LZWDico::CODE_DEPART 		= 261 ;	// premier code libre
const unsigned int 	 LZWDico::NOMBRE_BITS_MIN	= 9 ;		// longueur min de code
const unsigned int 	 LZWDico::NOMBRE_BITS_MAX	= 16 ;   // lg max absolue de code
																			// (tient encore dans un int)

LZWEntree::code LZWDico::premier( unsigned int nbits )
//
// Renvoie un nombre premier dpendant de la longueur actuelle du
// code, passe en argument. En effet, l'accs au dico comporte un
// calcul de hachage. Or, pour obtenir des performances correctes,
// la taille d'une hashtable doit tre :
// - un nombre premier ;
// - sensiblement suprieure au nombre maximum d'lments  stocker.
//
// Les premiers ci-dessous sont 23% suprieurs  la puissance de 2
// correspondant au nombre de bits, ce qui est juste suffisant selon
// mes constatations. (par exemple, pour 13 bits (max 8192 lments),
// la valeur choisie de 10103 conduit  un temps 5 fois meilleur que
// 9851, alors qu'aller au-del n'apporte plus aucune amlioration).
// (Nelson prconise une marge de 20%, moi j'ai trouv qu'il vaut en-
// la peine d'aller  23%, mais pas au-del ; je suppose que cela peut
// dpendre des environnements : mmoire, processeur, etc).
{
	switch( nbits )
	{
		case 9 :		return 641 ;
		case 10 :   return 1277 ;
		case 11 :   return 2531 ;
		case 12 :   return 5051 ;	// Nelson prconisait 5021
		case 13 :   return 10103  ;
		case 14 :   return 20219 ;
		case 15 :   return 40423;
	}
	assert( 0 ) ;	// fera tilt si nbits < 9 ou > 15
	return 0 ;		// jamais excut, mais calme le (stupide) compilateur
}


// Constructeur du dictionnaire LZW
// nombre_bits_final est la longueur de code au del de laquelle
// on souhaite vider le dictionnaire (tout le monde n'a pas 16MO
// de mmoire !)
// CODE_LIMITE est le dernier code pouvant tre mis avec la lon-
// gueur de code actuelle
// CODE_FINAL est le plus haut code que nous pourrons mettre avec
// la longueur de code nombre_bits_final (donc juste avant vidage)
// Il est crucial de faire un minimum de calculs  chaque octet lu,
// c'est pourquoi l'on stocke de nombreux paramtres (quitte  les
// remettre  jour ds que ncessaire).
//
// maxval est une fonction renvoyant le plus haut code possible pour
// une longueur donne (par exemple 511 pour 9 bits).
//
LZWDico::LZWDico( unsigned int nombre_bits_final )
		 : NOMBRE_BITS_ACTUEL( NOMBRE_BITS_MIN ),
			TAILLE_DICO( premier( nombre_bits_final ) ),
			DICO( new LZWEntree[ premier( nombre_bits_final ) ] ),
			CODE_FINAL( maxval( nombre_bits_final ) ),
			CODE_LIMITE( maxval( NOMBRE_BITS_MIN ) ),
			CODE_LIBRE_SUIVANT( CODE_DEPART )
{
	assert( 	nombre_bits_final >= NOMBRE_BITS_MIN
			&& nombre_bits_final <= NOMBRE_BITS_MAX
			&& DICO != 0 ) ;
}


LZWDico::~LZWDico()
{
	delete [] DICO ;
}


void LZWDico::vider()
//
// Rtablit le dico dans son tat de dpart (en mettant des RIEN dans
// ses entres, en mettant la longueur de code  celle de dpart, etc)
{
	NOMBRE_BITS_ACTUEL = NOMBRE_BITS_MIN ;
	CODE_LIBRE_SUIVANT = CODE_DEPART ;
	CODE_LIMITE = maxval( NOMBRE_BITS_MIN ) ;
	for( LZWEntree::code i = 0 ; i < TAILLE_DICO ; ++i )
		DICO[ i ].ajoute( LZWEntree::RIEN, LZWEntree::RIEN ) ;
}


LZWEntree::code LZWDico::index ( LZWEntree::code cparent, unsigned char carfils ) const
//
// Cette fonction reoit en argument :
// - le code (l'index dans le dico) cparent, dont on veut trouver le fils ;
// - le caractre qui tiquette le fils que l'on cherche.
// Elle doit renvoyer l'index du noeud fils dans le dictionnaire. Ce n'est
// pas trivial car il n'y a pas de pointeur dans le sens parent vers fils.
// Le principe est qu'un noeud fils est rang  un index calcul par hacha-
// ge  partir du code du parent et du caractre qui tiquette le fils.
{
	// Ceci est la fonction de hachage donnant l'index prsum du fils
	LZWEntree::code ix	= ( ( ( LZWEntree::code )carfils ) << ( nbits() - 8 ) ) ^ cparent ;

	// Il reste  rsoudre les ex-aequos  cet index (collisions)
	// Pour cela on parcourt circulairement le dictionnaire, jusqu'
	// trouver un noeud vide (on renvoie RIEN), ou un noeud dont les
	// lments code et car sont conformes  ceux en argument (rso-
	// lution russie de la collision). On renvoie alors l'index.
	LZWEntree::code dec  = ix ? tailleDico() - ix : 1 ;
	for( ; ; ix = ( ix >= dec ) ? ix - dec : ix + tailleDico() - dec )
	{
		if( DICO[ ix ].leCode() == LZWEntree::RIEN )
			break ;
		if( DICO[ ix ].leParent() == cparent && DICO[ ix ].leCar() == carfils )
			break ;
	}
	return ix ;
}


#ifdef DEPANNAGE
void LZWDico::lister() const
//
// J'avoue humblement que cela m'a servi durant la mise au point...
{
	cout << endl ;
	cout << "taille dico : " << TAILLE_DICO
		  << endl ;
	cout << "nbre bits :"
		  << "\tmin "		<< NOMBRE_BITS_MIN
		  << "\tmax "		<< NOMBRE_BITS_MAX
		  << "\tactuel "  << NOMBRE_BITS_ACTUEL
		  << endl ;
	cout << "codes :"
		  << "\tdpart "	<< CODE_DEPART
		  << "\tlibre "	<< CODE_LIBRE_SUIVANT
		  << "\tlimite "	<< CODE_LIMITE
		  << "\tfinal "	<< CODE_FINAL
		  << endl ;
}
#endif
