#ifndef LZWDICO_H
#define LZWDICO_H

//////////////////////////////////////////////////////////////////////////
// Ce fichier implmente le dictionnaire spcifique LZW dcrit par Mark
// Nelson dans son livre "La compression de donnes", chez Dunod qu'il me
// semble indispensable de consulter : je n'ai ni le courage ni le talent
// pour reprendre ici toutes les explications ; sachez que la suite n'est
// que la transposition en C++ de ce que vous trouvez en C classique dans
// Nelson. Par consquent, les algorythmes ont t recopis pratiquement
// tels quels, en tirant simplement parti des outils "objet" dvelopps
// pour simplifier et scuriser le code (donc faciliter la mise au point).
//
// -> Dans Nelson, voyez spcialement les pages 227  235, puis 240  248.
//
// Avant le dictionnaire (classe LZWico), il nous faut cependant dfinir
// une classes LZWEntree dcrivant la structure de chacune de ses entres.
// Chaque entre est un noeud tiquet par un caractre (donne prive CAR).
// La donne CODE est le code de la chane se terminant  ce noeud, tandis
// que CODE_PARENT est celui de la chane parente (la prcdente, sauf CAR).
// Pour remonter les noeuds, il suffit de suivre les codes parents qui sont
// des numros d'entres dans le dictionnaire ; en revanche pour descendre
// il faut raliser un hachage : voir la fonction index dans le .cpp.
//////////////////////////////////////////////////////////////////////////


///////////////
class LZWEntree {
///////////////
public :

typedef	unsigned int	code ;				// un alias pour la clart

static	const code		RIEN ;
static	const code		FIN_FICHIER ;
static	const code		ALLONGER_CODE ;
static	const code		VIDER_DICO ;
static	const code		IGNORER_DICO ;		// inutilis  ce jour (une autre
														// stratgie suite au dbordement
														// du dico, serait de l'ignorer en
														// ne comprimant plus, au lieu de
														// le vider ; j'avais prvu ce code
														//  tout hasard)

								// Constructeur
								LZWEntree() : CODE( RIEN ), CODE_PARENT( RIEN )
								{ }			  // CAR est sans importance ici

								// Mise  jour ultrieure de l'entre
			void				ajoute	( code c, code cparent,
											  unsigned char car = ( unsigned char )0 )
								{ CODE = c ; CODE_PARENT = cparent ; CAR = car ; }

								// Consultation des donnes de l'entre
			code				leCode	() const { return CODE ; }
			code				leParent	() const { return CODE_PARENT ; }
			unsigned char	leCar		() const { return CAR ; }

private :
			code				CODE, CODE_PARENT ;
			unsigned char	CAR ;
} ;


/////////////
class LZWDico {
/////////////
public :
											// Constructeur et destructeur
											LZWDico		( unsigned int nombre_bits_final ) ;
										  ~LZWDico		() ;

			// Consultation de donnes internes
			unsigned int				nbits			() const { return NOMBRE_BITS_ACTUEL ; }
			LZWEntree::code			tailleDico	() const { return TAILLE_DICO ; }

			// Renvoie une rfrence  la ime entre (sans test sur i !)
			LZWEntree &					operator[]	( LZWEntree::code i ) { return DICO[ i ] ; }

			LZWEntree::code			codeSuivant	() const { return CODE_LIBRE_SUIVANT ; }
static	LZWEntree::code			codeDepart	() 		{ return CODE_DEPART ; }

			enum							etatCode		{ ras, pluslong, troplong } ;
			etatCode						incrementer	() ;
			etatCode						incrementer	( unsigned int & ancien_nombre_bits ) ;
			void							allongerCode()	{ NOMBRE_BITS_ACTUEL++ ; }

static   LZWEntree::code			maxval	( unsigned int nbits )	{ return ( LZWEntree::code )( ( ( ( unsigned long ) 1 ) << nbits ) - 1 ) ; }
			LZWEntree::code			index			( LZWEntree::code cparent, unsigned char carfils ) const ;

			void							vider			() ;

#ifdef DEPANNAGE
			void							lister		() const ;
#endif


private :

static	const unsigned int		NOMBRE_BITS_MIN ;    // voir le .cpp
static	const unsigned int		NOMBRE_BITS_MAX ;
			unsigned int				NOMBRE_BITS_ACTUEL ;
static 	const LZWEntree::code	CODE_DEPART ;
			LZWEntree::code			CODE_LIMITE ;
			const LZWEntree::code	CODE_FINAL ;
			LZWEntree::code			CODE_LIBRE_SUIVANT ;
			LZWEntree *					DICO ;					// tableau des entres
			const LZWEntree::code	TAILLE_DICO ;

static   LZWEntree::code			premier	( unsigned int nbits ) ;

} ;


inline LZWDico::etatCode LZWDico::incrementer()
//
// Incrmente inconditionnellement la valeur du pro-
// chain code disponible, sans autre effet de bord.
// (Ne sert que pour la dcompression)
{
	CODE_LIBRE_SUIVANT++ ;
	return ras ;
}


inline LZWDico::etatCode LZWDico::incrementer( unsigned int & ancien_nombre_bits )
//
// Incrmente le prochain code libre, mais avec des effets de bord en cascade :
// - si la longueur de code est puise, il faut l'augmenter ;
// - si elle ne peut plus tre augmente, il faut vider le dico.
// Dans tous les cas, on renvoie une valeur de type numr ren-
// dant compte de que qu'on a fait ; quand un code spcial devra
// tre mis pour renseigner le dcompresseur, il devra l'tre
// avec l'ancienne longueur (qu'on aura peut-tre change ici m-
// me), donc on la passe  l'appelant dans un argument rfrence.
// Cette fonction sert pendant la compression.
{
	ancien_nombre_bits = NOMBRE_BITS_ACTUEL ;
	CODE_LIBRE_SUIVANT++ ;

	// Dbordement irrmdiable : la longueur de code ne peut tre augmente
	if( CODE_LIBRE_SUIVANT > CODE_FINAL )
	{
		vider() ;            // rtablit la longueur de code initiale
		return troplong ;		// avis de vidage
	}

	// Saturation de la longueur de code, qu'il suffit d'augmenter
	if( CODE_LIBRE_SUIVANT > CODE_LIMITE )
	{
		NOMBRE_BITS_ACTUEL++ ;
		CODE_LIMITE = maxval( NOMBRE_BITS_ACTUEL ) ;
		return pluslong ;		// avis d'incrmentation de la longueur
	}

	return ras ;				// rien  signaler (pas de dbordement)
}


#endif
