///////////////////////////////////////////////////////////////////////////
// Fichier inclu dans chacun des programmes principaux comp.cpp, dcompr.cpp
// Il s'agit essentiellement de paramtrer au mieux le traitement (compres-
// sion ou dcompression) du fichier en ligne de commande, ou des fichiers
// trouvs dans le fichier d'ordre (commenant par @) en ligne de commande.
///////////////////////////////////////////////////////////////////////////




// Traiter un fichier, c'est lui appliquer l'une des fonctions membres
// compresse ou decompresse de la classe LZW. Comme elles ont la mme
// signature, il est commode de gnraliser le traitement, en dfinis-
// sant un type de pointeur sur une telle fonction membre :

typedef LZW::erreur ( LZW::*LZW_membre ) ( const char *, const char * ) ;


// Cas simple : traiter un fichier. L'opration (compression ou dcompres-
// sion) est passe en argument 1 comme un pointeur sur l'une des 2 fonc-
// tions membres de la classe LZW.

unsigned traite_un( LZW_membre op, const char * source, const char * cible )
{
	LZW lzw( NBITS ) ;	// construit une classe LZW
	LZW::erreur er = ( lzw.*op )( source, cible ) ;	// (d)compression
	if( er != LZW::ras ) // affichages en sortie standard si erreur
		cout << PETIT_NOM << " : " << lzw.message( er ) << "." << endl ;
	return er != LZW::ras && er != LZW::taille ;	// retour 1 si erreur grave
}


// Traiter plusieurs fichiers, c'est plusieurs fois en traiter un...
// L'argument commandes est le nom du fichier d'ordre contenant la
// liste des fichiers  traiter.

unsigned traite_plusieurs( LZW_membre op, const char * commandes )
{
	ifstream c( commandes ) ;	// tente d'ouvrir fichier de commandes
	if( ! c.good() )
	{
		cout << PETIT_NOM << " : " << "impossible d'ouvrir le fichier de commandes." << endl ;
		return 2 ;
	}
	unsigned compte = 0 ;
	unsigned retour = 0 ;
	while( ! c.eof() )	// parcours du fichier de commandes
	{
		// On y lit les noms, en sautant les lignes vides (donc permises)
		// et en mettant un message en cas d'erreur manifeste
		char ligne[ MAXPATH ] ;
		c.getline( ligne, MAXPATH ) ;	++compte ;	// lecture ligne
		if( ! ligne || ! *ligne )                 // elle est vide
			continue ;
		char * arg1 = strtok( ligne, " \t" ) ;    // arg1 et arg2 sont obligatoires
		char * arg2 = strtok( NULL, " \t" ) ;     // ne doivent pas commencer par @
		if( !arg1 || arg1[ 0 ] == '@' || !arg2 || arg2[ 0 ] == '@' )
		{
			cout << PETIT_NOM << " : " << "ligne de commande " << compte << " erronne." << endl ;
			retour = 1 ; continue ;
		}
		retour |= traite_un( op, arg1, arg2 ) ;	// traitement de chaque fichier lu
	}
	return retour ;
}


void usage()
//
// Baratin mis en sortie standard en cas d'erreur ou ommission d'arguments
{
	cout << endl << GRAND_NOM << " LZW  code variable de 9  " << NBITS << " bits." << endl
					 << "Usage :	" << PETIT_NOM << "	chemin_source	chemin_cible" << endl
					 << "   ou :	" << PETIT_NOM << "	@fichier_commandes" << endl ;
}
